import pyawr.mwoffice as mwo
from tkinter import messagebox
import warnings
#
#
#*****************************************************************************
#
class _DataFiles():
    '''
    Methods for collection of Data Files in the AWRDE project
    
     Parameters
    ----------
    awrde: The AWRDE object returned from awrde_utils.establish_link()
    
    '''
    def __init__(self, awrde):#----------------------------------------
        self.awrde = awrde
        #
    def _build_data_files_dict(self):#--------------------------------
        '''Create list of all graph names in project
    
        Parameters
        ----------
        None
    
        Returns
        -------
        graphs_dict: dict[element_objects]
                         Each element in the list is a graph name
        '''        
        self._data_files_dict = {}
        for df in self.awrde.Project.DataFiles:
            data_file = _DataFile(self.awrde, df.Name)
            self._data_files_dict[df.Name] = data_file
        #end for
        #
    @property
    def data_files_dict(self) -> dict:#-----------------------------------------------
        '''return data file dictionary'''
        self._build_data_files_dict()
        return self._data_files_dict
        #
    @property
    def data_file_names_list(self) -> list[str]:#-------------------------------------------
        '''returns list of data file names'''
        data_file_name_list = list()
        for df in self.awrde.Project.DataFiles:
            data_file_name_list.append(df.Name)
        #end for
        return data_file_name_list
        #
    def add_dc_iv_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a DC-IV file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'DC-IV')
        #
    def add_dscr_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a DSCR file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'DSCR')
        #
    def add_gmdif_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a Generalized MDIF file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'GMDIF')
        #
    def add_gmdif_nport_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a Generalized MDIF N-Port file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'GMDIF N-Port')
        #
    def add_mdif_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a MDIF file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'MDIF')
        #
    def add_raw_port_parameters_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a Raw Port Parameters file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'Raw Port Parameters')
        #
    def add_text_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a Text file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'Text')
        #
    def add_touchstone_file(self, data_file_name: str):#-----------------------------------
        '''
        Add a Touchstone file
        
        Parameters
        ----------
        data_file_name: string
        '''
        self._add_data_file(data_file_name, 'Touchstone')
        #
    def remove_data_file(self, data_file_name: str) -> bool:
        '''
        Delete a data file from the AWRDE project
        
        Parameters
        ----------
        data_file_name: string,
                 name of the Data File to be deleted
        
        Returns
        -------
        data_file_removed: bool
                    True if data file successfully deleted,
                    False if data file could not be deleted
                    
        '''        
        if not isinstance(data_file_name, str):
            raise TypeError('data_file_name must be a string type')
        #end if
        data_file_removed = False               
        try:
            for df_idx in range(self.awrde.Project.DataFiles.Count):
                if self.awrde.Project.DataFiles[df_idx].Name == data_file_name:
                    self.awrde.Project.DataFiles.Remove(df_idx+1)
                    data_file_removed = True
                    break
                #end if
            #end for
        except:
            warnings.warn('remove_data_file: Data File ' + data_file_name + ' did not get removed')
        #end for
        return data_file_removed
        #
    def rename_data_file(self, data_file_name: str, new_data_file_name: str):#--------------------
        '''
        rename Data File
    
        Parameters
        ----------
        data_file_name: string
                         existing graph name
        new_data_file_name: string
                         new graph name
        '''        
        if not isinstance(data_file_name, str):
            raise TypeError('data_file_name must be a string')
        #end if
        if not isinstance(new_data_file_name, str):
            raise TypeError('new_data_file_name must be a string')
        #end if
        data_file_name_list = self.data_file_names_list
        found_it = False
        for data_file_name_from_list in data_file_name_list:
            if data_file_name_from_list == data_file_name:
                found_it = True
                if new_data_file_name in data_file_name_list:
                    YesNo = messagebox.askyesno('Rename Data File','Data File '\
                        + new_data_file_name + ' exists. Remove existing Data File ?')
                    if YesNo:
                        self.remove_data_file(new_data_file_name)
                    else:
                        new_data_file_name += ' 1'
                    #end if
                #end if
                try:
                    df = self.awrde.Project.DataFiles(data_file_name)
                    df.Name = new_data_file_name
                except:
                    raise RuntimeError('problem with renaming data file')
                #end try
            #end if
        #end for
        if not found_it:
            warnings.warn('data file rename failed: ' + data_file_name)
        #end if
        #
    def copy_data_file(self, data_file_name: str, new_data_file_name: str):#-----------------------
        '''
        copy Data File
    
        Parameters
        ----------
        data_file_name: string
                         existing data file name
        new_data_file_name: string
                         new data file name
        '''
        if not isinstance(data_file_name, str):
            raise TypeError('data_file_name must be a string')
        #end if
        if not isinstance(new_data_file_name, str):
            raise TypeError('new_data)file_name must be a string')
        #end if        
        for df in self.awrde.Project.DataFiles:
            if df.Name == new_data_file_name:
                new_data_file_name = new_data_file_name + ' 1'
            #end if
        #end for
        #
        try:
            for df_idx in range(1, self.awrde.Project.DataFiles.Count+1):
                df = self.awrde.Project.DataFiles(df_idx)
                if df.Name == data_file_name:
                    self.awrde.Project.DataFiles.Copy(df_idx, new_data_file_name)
                #end if
            #end for
        except:
            raise RuntimeError('problem with copying data file')
        #end try
        #
    def import_data_file(self, data_file_name: str, file_path: str, data_file_type: str):#------
        '''
        import data file
        
        Parameters
        ----------
        data_file_name: string
               name of the data file as it will appear in the project
               
        file_path: string
               full directory and file name of the file to be imported
               
        data_file_type: string
               valid values:
                    DC-IV File
                    DSCR File
                    Generalized MDIF Data File
                    Generalized MDIF N-Port File
                    MDIF File
                    Raw Port Parameters File
                    Text File
                    Touchstone File
        '''
        if not isinstance(data_file_name, str):
            raise TypeError('data_file_name must be string type')
        #end if
        if not isinstance(file_path, str):
            raise TypeError('file_path must be string type')
        #end if
        if not isinstance(data_file_type, str):
            raise TypeError('data_file_type must be string type')
        #end if
        #
        if data_file_type == 'DC-IV File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_IV
        elif data_file_type == 'DSCR File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_DSCR
        elif data_file_type == 'Generalized MDIF Data File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_GMDIF
        elif data_file_type == 'Generalized MDIF N-Port File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_GMDIFD
        elif data_file_type == 'MDIF File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_MDIF
        elif data_file_type == 'Raw Port Parameters File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_RAW
        elif data_file_type == 'Text File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_TXT
        elif data_file_type == 'Touchstone File':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_SNP
        else:
            raise RuntimeError('data_file_type not recognized: ' + data_file_type)
        #end if
        if self._does_data_file_exist(data_file_name):
            YesNo = messagebox.askyesno('Import Data File','Data File ' + data_file_name + ' exists. Delete existing Data File?')
            if YesNo:
                self.remove_data_file(data_file_name)
            #end if
        #end if
        try:
            self.awrde.Project.DataFiles.Import(data_file_name, file_path, data_file_type_enum)
        except:
            raise RuntimeError('Problem with importing data file')
        #end try
        #
    def _does_data_file_exist(self, data_file_name: str) -> bool:#------------------------
        data_file_exists = False
        for df in self.awrde.Project.DataFiles:
            if df.Name == data_file_name:
                data_file_exists = True
            #end if
        #end for
        return data_file_exists
        #
    def _add_data_file(self, data_file_name, data_file_type):#----------------------------
        if data_file_type == 'DC-IV':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_IV
        elif data_file_type == 'DSCR':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_DSCR
        elif data_file_type == 'GMDIF':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_GMDIF
        elif data_file_type == 'GMDIF N-Port':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_GMDIFD
        elif data_file_type == 'MDIF':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_MDIF
        elif data_file_type == 'Raw Port Parameters':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_RAW
        elif data_file_type == 'Text':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_TXT
        elif data_file_type == 'Touchstone':
            data_file_type_enum = mwo.mwDataFileType.mwDFT_SNP
        #end if
        if self._does_data_file_exist(data_file_name):
            YesNo = messagebox.askyesno('Add Data File','Data File ' + data_file_name + ' exists. Delete existing Data File?')
            if YesNo:
                self.remove_data_file(data_file_name)
            #end if
        #end if
        self.awrde.Project.DataFiles.AddNew(data_file_name, data_file_type_enum)
#
#*****************************************************************************
#
class _DataFile():
    def __init__(self, awrde, data_file_name: str):#-----------------------------------
        self.awrde = awrde
        self._initialize_data_file(data_file_name)
        #
    def _initialize_data_file(self, data_file_name: str):#-------------------------------------
        try:
            self._df = self.awrde.Project.DataFiles(data_file_name)
        except:
            raise RuntimeError('Data File does not exist: ' + data_file_name)
        #end try
        #
    @property
    def data_file_name(self) -> str:#------------------------------------------------------
        '''Returns name of the Data File'''
        return self._df.Name
        #
    @property
    def data_file_type(self) -> str:#--------------------------------------------------
        '''
        Reports data file type 
        
        Returns
        -------
        data_file_type_str: string
              DC-IV File, DSCR File, Generalized MDIF Data File, Generalized MDIF N-Port File,
              MDIF File, Raw Port Parameters File, Text File, Touchstone File
        
        ''' 
        data_file_type_enum = self._df.Type
        if mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_IV:
            data_file_type_str = 'DC-IV File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_DSCR:
            data_file_type_str = 'DSCR File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_GMDIF:
            data_file_type_str = 'Generalized MDIF Data File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_GMDIFD:
            data_file_type_str = 'Generalized MDIF N-Port File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_MDIF:
            data_file_type_str = 'MDIF File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_RAW:
            data_file_type_str = 'Raw Port Parameters File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_TXT:
            data_file_type_str = 'Text File'
        elif mwo.mwDataFileType(data_file_type_enum) == mwo.mwDataFileType.mwDFT_SNP:
            data_file_type_str = 'Touchstone File'
        #end if
        return data_file_type_str
        #
    def export_data_file(self, file_path: str) -> bool:#--------------------------------------
        '''
        exports the data file to the directory and name specified by file_path
        
        Parameters
        ----------
        file_path: string
              full director and file name
        
        Returns
        -------
        export_successful: bool
        '''
        if not isinstance(file_path,str):
            raise TypeError('file_path must be string type')
        #end if
        #
        export_successful = True
        try:
            self._df.Export(file_path)
        except:
            export_successful = False
        #end try
        return export_successful
        
#
#*****************************************************************************
#


from pyawr_utils import common_utils, awrde_utils
import numpy as np
import warnings
import itertools

_Elements = common_utils._Elements
_Equations = common_utils._Equations
#
#*****************************************************************************************************
#

# class _Options():

class _Schematics():
    '''Class that contains class constructors and methods at the Circuit Schematics level in the AWR project tree'''

    def __init__(self, project):
        self._schematics = project.Schematics
        self._build_circuit_schematics_dict()

    def _build_circuit_schematics_dict(self):
        self._circuit_schematics_dict = {}
        for schematic in self._schematics:
            try:
                self._circuit_schematics_dict[schematic.Name] = _Schematic(self._schematics(schematic.Name))

            except:
                raise RuntimeError('Failed to create circuit schematics dictionary')

    @property
    def circuit_schematics_dict(self) -> dict:
        return self._circuit_schematics_dict

    def add_circuit_schematic(self, circuit_schematic_name: str) -> bool:
        if not isinstance(circuit_schematic_name, str):
            raise TypeError('Schematic diagram name must be a string type')

        elif isinstance(circuit_schematic_name, str):
            if circuit_schematic_name == self._schematics.GetUniqueName(circuit_schematic_name):
                new_circuit_schematic = self._schematics.Add(circuit_schematic_name)
                self._circuit_schematics_dict[new_circuit_schematic.Name] = _Schematic(
                                                                            self._schematics(circuit_schematic_name))

            elif circuit_schematic_name != self._schematics.GetUniqueName(circuit_schematic_name):
                warnings.warn('Circuit schematic name ' + circuit_schematic_name \
                               + ' is already in use')

            else:
                warnings.warn('Circuit schematic failed to be added')

        return self._circuit_schematics_dict

    def remove_circuit_schematic(self, circuit_schematic_name: str) -> dict:
        if not isinstance(circuit_schematic_name, str):
            raise TypeError('Schematic diagram name must be a string type')

        elif isinstance(circuit_schematic_name, str):
            try:
                for circuit_schematic_index in range(1, self._schematics.Count + 1):
                    if circuit_schematic_name == self._schematics(circuit_schematic_index).Name:
                        circuit_schematic_removed = self._schematics.Remove(circuit_schematic_index)

                        if circuit_schematic_removed == True:
                            self._circuit_schematics_dict.pop(circuit_schematic_name)
                        else:
                            warnings.warn('Circuit schematic ' + new_circuit_schematic_name \
                                          + 'was not deleted')

                        break

            except:
                warnings.warn('remove_circuit_schematic: Circuit Schematic ' + circuit_schematic_name  \
                              + ' did not get removed')

        return self._circuit_schematics_dict

    def copy_circuit_schematic(self, existing_circuit_schematic_name: str, new_circuit_schematic_name: str) -> dict:
        if not isinstance(existing_circuit_schematic_name, str) and not isinstance(new_circuit_schematic_name, str):
            raise TypeError('Schematic diagram names must be a string type')

        elif new_circuit_schematic_name != self._schematics.GetUniqueName(new_circuit_schematic_name):
            warnings.warn('New circuit schematic name ' + new_circuit_schematic_name \
                                + 'is already in use')

        elif self._circuit_schematics_dict.get(existing_circuit_schematic_name) is None:
            warnings.warn('Circuit schematic name ' + existing_circuit_schematic_name \
                          + 'cannot be copied because it does not exist')

        elif existing_circuit_schematic_name in self._circuit_schematics_dict.keys():
            schematic_count = 0
            for schematic in self._schematics:
                if existing_circuit_schematic_name == schematic.Name:
                    new_circuit_schematic = self._schematics.Copy(schematic_count, new_circuit_schematic_name)
                    self._circuit_schematics_dict[new_circuit_schematic.Name] = _Schematic(
                                                                        self._schematics(new_circuit_schematic_name))
                    break

                schematic_count += 1

        return self._circuit_schematics_dict

    def rename_circuit_schematic(self, existing_circuit_schematic_name: str, new_circuit_schematic_name: str) -> dict:
        if not isinstance(existing_circuit_schematic_name, str) and not isinstance(new_circuit_schematic_name, str):
            raise TypeError('Schematic diagram names must be a string type')

        elif new_circuit_schematic_name != self._schematics.GetUniqueName(new_circuit_schematic_name):
            warnings.warn('New circuit schematic name ' + new_circuit_schematic_name \
                                + 'is already in use')

        elif existing_circuit_schematic_name in self._circuit_schematics_dict.keys():
            schematic_count = 0
            for schematic in self._schematics:
                if existing_circuit_schematic_name == schematic.Name:
                    self._schematics(existing_circuit_schematic_name).Name = new_circuit_schematic_name
                    self._circuit_schematics_dict[new_circuit_schematic_name] = self._circuit_schematics_dict[
                                                                                    existing_circuit_schematic_name]
                    self._circuit_schematics_dict.pop(existing_circuit_schematic_name)
                    break

                schematic_count += 1

        return self._circuit_schematics_dict
#
#*****************************************************************************************************
#
class _Schematic(_Elements): 

    def __init__(self, circuit_schematic):
        self._circuit_schematic = circuit_schematic
        _Elements.__init__(self, self._circuit_schematic, 'Schematic')
        #_Equations.__init__(self, self._circuit_schematic)
        

    @property
    def circuit_schematic_name(self) -> str:
        '''Returns name of circuit schematics'''
        return self._circuit_schematic.Name

    @property
    def frequency_range(self) -> np.ndarray:
        '''Returns circuit schematic frequency range'''
        self.frequency_array = np.array([round(frequency.Value) for frequency in self._circuit_schematic.Frequencies])
        return self.frequency_array

    def set_frequency_range(self, frequency_array: np.ndarray, units_str: str = 'GHz'):
        '''Sets circuit cchematic frequency range'''
        units_dict = {'Hz':1, 'kHz':1e3, 'MHz':1e6, 'GHz':1e9}
        try:
            frequency_multiplier = units_dict[units_str]

        except:
            raise RuntimeError('Incorrect units: ' + units_str)

        frequency_array = np.multiply(frequency_array, frequency_multiplier)
        try:
            self._circuit_schematic.UseProjectFrequencies = False
            self._circuit_schematic.Frequencies.Clear()
            self._circuit_schematic.Frequencies.AddMultiple(frequency_array)

        except:
            raise RuntimeError('Error in setting circuit schematic frequencies')

    @property
    def use_project_frequencies(self) -> bool:
        return self._circuit_schematic.UseProjectFrequencies

    @use_project_frequencies.setter
    def use_project_frequencies(self, set_state: bool) -> bool:
        if not isinstance(set_state, bool):
            raise TypeError('set_state must be a boolean')

        self._circuit_schematic.UseProjectFrequencies = set_state
        return self._circuit_schematic.UseProjectFrequencies

    @property
    def options(self)->dict:
        '''Gets Circuit Schematic settings with a dictionary of option values'''
        pass


    @options.setter
    def options(self, circuit_schematic_option_set_dict: dict):
        '''Sets Circuit Schematic settings with a dictionary of option values'''
        for option_set_dict_key, option_set_dict_value in circuit_schematic_option_set_dict.items():
            pass





































import pyawr.mwoffice as mwo
from pyawr_utils import vss_utils
from pyawr_utils import common_utils
from tkinter import messagebox
import warnings
import numpy as np
#
_SystemDiagrams = vss_utils._SystemDiagrams
_ProcessDefinitions = common_utils._ProcessDefinitions
#
#*****************************************************************************
#
class _Graphs():
    '''
    Methods for the collection of Graphs in the AWRDE project
    
    Parameters
    ----------
    awrde: The AWRDE object returned from awrde_utils.establish_link()

    '''    
    def __init__(self, awrde):#-----------------------------------------------------------
        self.awrde = awrde
        self._build_graphs_dict()
        #
    def _build_graphs_dict(self) -> dict:#--------------------------------------------------
        '''Create list of all graph names in project
    
        Parameters
        ----------
        None
    
        Returns
        -------
        graphs_dict: dict[element_objects]
                         Each element in the list is a graph name
        '''
        self._graphs_dict = {}
        for gr in self.awrde.Project.Graphs:
            graph = _Graph(self.awrde, gr.Name)
            self._graphs_dict[graph.graph_name] = graph
        #end for
        #
    @property
    def graph_dict(self) -> dict:#---------------------------------------------------------
        '''returns graph dictionary'''
        self._build_graphs_dict()
        return self._graphs_dict
        #    
    @property
    def graph_name_list(self) -> list[str]:#----------------------------------------------
        '''
        Creates list of all graph names in project
        
        Parameters
        ----------
        None
        
        Returns
        -------
        graph_name_list: list[string]
                   Each item in the list is a graph name
        '''
        graph_name_list = list()
        for gr in self.awrde.Project.Graphs:
            graph_name_list.append(gr.Name)
        #end for
        return graph_name_list
        #
    def add_rectangular_graph(self, graph_name: str) -> object:#-----------------------------
        '''
        Add a rectangular graph
        
        Parameters
        ----------
        graph_name: string,
                 name of the Graph to be added
             
        '''
        self._add_graph(graph_name, 'Rectangular')
        #
    def add_rectangular_real_imag_graph(self, graph_name: str) -> object:
        '''
        Add a Rectangular - Real/Imag graph
        
        Parameters
        ----------
        graph_name: string,
                 name of the Graph to be added
        
        '''
        self._add_graph(graph_name, 'Rect_RealImag')
        #    
    def add_smith_chart_graph(self, graph_name: str) -> object:#-----------------------------
        '''
        Add a Smith Chart graph
        
        Parameters
        ----------
        graph_name: string,
                 name of the Graph to be added
             
        '''
        self._add_graph(graph_name, 'Smith')
        #
    def add_polar_graph(self, graph_name: str) -> object:#------------------------------------
        '''
        Add a Polar graph
        
        Parameters
        ----------
        graph_name: string,
                 name of the Graph to be added
             
        '''
        self._add_graph(graph_name, 'Polar')
        #
    def add_histogram_graph(self, graph_name: str) -> object:#-------------------------------
        '''
        Add a Histogram graph
        
        Parameters
        ----------
        graph_name: string,
                 name of the Graph to be added
             
        '''
        self._add_graph(graph_name, 'Histogram')
        #
    def add_antenna_plot_graph(self, graph_name: str) -> object:#------------------------------
        '''
        Add an Antenna Plot graph
        
        Parameters
        ----------
        graph_name: string,
                 name of the Graph to be added
             
        '''
        self._add_graph(graph_name, 'Antenna')
        #
    def add_tabular_graph(self, graph_name: str) -> object:#---------------------------------
        '''
        Add a Tabular graph
        
        Parameters
        ----------
        graph_name: string,
                 name of the Graph to be added
        
           '''
        self._add_graph(graph_name, 'Tabular')
        #
    def add_constellation_graph(self, graph_name: str) -> object:#-----------------------------
        '''
        Add a Constellation graph
        
        Parameters
        ----------
        graph_name: string,
                 name of the Graph to be added
             
        '''
        self._add_graph(graph_name, 'Constellation')
        #
    def add_3D_graph(self, graph_name: str) -> object: #-------------------------------------
        '''
        Add a 3D graph
        
        Parameters
        ----------
        graph_name: string,
                 name of the Graph to be added
            
        '''
        self._add_graph(graph_name, '3D')
        #    
    def remove_graph(self, graph_name: str) -> bool:#--------------------------------------------
        '''
        Delete a graph from the AWRDE project
        
        Parameters
        ----------
        graph_name: string,
                 name of the Graph to be deleted
        
        Returns
        -------
        graph_removed: bool
                    True if graph successfully deleted, False if graph could not be deleted
                    
        '''
        if not isinstance(graph_name, str):
            raise TypeError('graph_name must be a string type')
        #end if
        graph_removed = False               
        try:
            for gr_idx in range(self.awrde.Project.Graphs.Count):
                if self.awrde.Project.Graphs[gr_idx].Name == graph_name:
                    self.awrde.Project.Graphs.Remove(gr_idx+1)
                    graph_removed = True
                    break
                #end if
            #end for
        except:
            warnings.warn('remove_graph: Graph ' + graph_name + ' did not get removed')
        #end for
        return graph_removed
        #
    def rename_graph(self, graph_name: str, new_graph_name: str):#--------------------
        '''
        rename Graph
    
        Parameters
        ----------
        graph_name: string
                         existing graph name
        new_graph_name: string
                         new graph name
        '''        
        if not isinstance(graph_name, str):
            raise TypeError('graph_name must be a string')
        #end if
        if not isinstance(new_graph_name, str):
            raise TypeError('new_graph_name must be a string')
        #end if
        graph_name_list = self.graph_name_list
        found_it = False
        for graph_name_from_list in graph_name_list:
            if graph_name_from_list == graph_name:
                found_it = True
                if new_graph_name in graph_name_list:
                    YesNo = messagebox.askyesno('Rename Graph','Graph '\
                        + new_graph_name + ' exists. Remove existing Graph ?')
                    if YesNo:
                        self.remove_graph(new_graph_name)
                    else:
                        new_graph_name += ' 1'
                    #end if
                #end if
                try:
                    gr = self.awrde.Project.Graphs(graph_name)
                    gr.Name = new_graph_name
                except:
                    raise RuntimeError('problem with renaming graph')
                #end try
            #end if
        #end for
        if not found_it:
            warnings.warn('graph rename failed: ' + graph_name)
        #end if
        #
    def copy_graph(self, graph_name: str, new_graph_name: str):#-----------------------
        '''
        copy Graph
    
        Parameters
        ----------
        graph_name: string
                         existing graph name
        new_graph_name: string
                         new graph name
        '''
        if not isinstance(graph_name, str):
            raise TypeError('graph_name must be a string')
        #end if
        if not isinstance(new_graph_name, str):
            raise TypeError('new_graph_name must be a string')
        #end if        
        for gr in self.awrde.Project.Graphs:
            if gr.Name == new_graph_name:
                new_graph_name = new_graph_name + ' 1'
            #end if
        #end for
        #
        try:
            for gr_idx in range(1, self.awrde.Project.Graphs.Count+1):
                gr = self.awrde.Project.Graphs(gr_idx)
                if gr.Name == graph_name:
                    self.awrde.Project.Graphs.Copy(gr_idx, new_graph_name)
                    graph_copied = True
                #end if
            #end for
        except:
            raise RuntimeError('problem with copying graph')
        #end try
        #        
    def _does_graph_exist(self, graph_name: str) -> bool:#---------------------------------------
        graph_exists = False
        for gr in self.awrde.Project.Graphs:
            if gr.Name == graph_name:
                graph_exists = True
            #end if
        #end for
        return graph_exists
        #
    def _add_graph(self, graph_name, graph_type):
        if graph_type == 'Rectangular':
            graph_type_enum = mwo.mwGraphType.mwGT_Rectangular
        elif graph_type == 'Rect_RealImag':
            graph_type_enum = mwo.mwGraphType.mwGT_RectangularComplex
        elif graph_type == 'Smith':
            graph_type_enum = mwo.mwGraphType.mwGT_SmithChart
        elif graph_type == 'Polar':
            graph_type_enum = mwo.mwGraphType.mwGT_Polar
        elif graph_type == 'Histogram':
            graph_type_enum = mwo.mwGraphType.mwGT_Histogram
        elif graph_type == 'Antenna':
            graph_type_enum = mwo.mwGraphType.mwGT_Antenna
        elif graph_type == 'Tabular':
            graph_type_enum = mwo.mwGraphType.mwGT_Tabular
        elif graph_type == 'Constellation':
            graph_type_enum = mwo.mwGraphType.mwGT_Constellation
        elif graph_type == '3D':
            graph_type_enum = mwo.mwGraphType.mwGT_ThreeDim
        else:
            raise RuntimeError(graph_type + ' not recognized')
        #end if
        #
        if type(graph_name) != str:
            raise RuntimeError('graph_name must be a string type')
        #end if
        if self._does_graph_exist(graph_name):
            YesNo = messagebox.askyesno('AddGraph','Graph ' + graph_name + ' exists. Delete existing Graph?')
            if YesNo:
                self.remove_graph(graph_name)
            #end if
        #end if
        #
        self.awrde.Project.Graphs.Add(graph_name, graph_type_enum)       
        #    

#
#*****************************************************************************
#
class _Measurements():
    '''
    Methods for collection of measurements within a graph
    
    Parameters
    ----------
    awrde: object variable
         The AWRDE object returned from awrde_utils.establish_link()
    
    graph_name: string
         Name of the graph that the measurement belongs to
         
    built_dict: bool, optional
           dtermines whether _build_measurements_dict() is run at initialization
    '''    
    def __init__(self, awrde, graph_name: str):#---------------------------
        self.awrde = awrde
        self._initialize_mesaurements(graph_name)
        self._build_measurements_dict()
        #
    def _initialize_mesaurements(self, graph_name):#-------------------------------------
        if not isinstance(graph_name, str):
            raise TypeError('graph_name must be a string')
        #end if
        try:
            self._gr = self.awrde.Project.Graphs(graph_name)
        except:
            raise RuntimeError('Graph not recognized ' + graph_name)
        #end try
        self._graph_name = graph_name
        #
    def _build_measurements_dict(self):#--------------------------------------------------------
        '''
        creates a dictionary of measurements in a graph
        '''
        self._graphs_dict = {}
        meas_idx = 0
        for meas in self._gr.Measurements:
            measurement = _Measurement(self.awrde, self._gr.Name, meas.Name)
            self._graphs_dict[meas_idx] = measurement
            meas_idx += 1
        #end for
        #
    @property
    def measurements_dict(self) -> dict:#---------------------------------------------------
        '''returns measurements dictionary'''
        self._build_measurements_dict()
        return self._graphs_dict
        #
    @property
    def measurement_name_list(self) -> list[str]:#-----------------------------------------
        '''Returns list of measurement names in the graph'''
        measurement_name_list = list()
        for meas in self._gr.Measurements:
            measurement_name_list.append(meas.Name)
        #end for
        return measurement_name_list
        #
    def add_measurement(self, source_doc: str, measurement: str):#------------------------
        '''
        Add measurement to graph
        
        Parameters
        ----------
        Source Doc: string
                name of the source document: Circuit schematic name, Systeme diagram name,
                Data File name, EM doc name
        measurement: string
                The measurement as a string
        '''
        try:
            self._gr.Measurements.Add(source_doc, measurement)
        except:
            raise RuntimeError('Invalid measurement add: ' + source_doc + ' , ' + measurement)
        #end try
        #
    def remove_measurement(self, measurement_name: str) -> bool:#-----------------------------
        '''
        Delete a measurement from the graph
        
        Parameters
        ----------
        measurement_name: string
              The measurement name in format source_doc:measurement
              
        Returns
        -------
        measurement_removed: bool
              True if measurement successfully removed
        '''
        measurement_removed = False
        try:
            if self._gr.Measurements.Exists(measurement_name):
                self._gr.Measurements.Remove(measurement_name)
                measurement_removed = True
            #end if
        except:
            measurement_removed = False
            warnings.warn('remove_measurement: measurement ' + measurement_name + ' did not get removed')
        #end try
        return measurement_removed
        #    
#
#*****************************************************************************
#
class _Measurement(_ProcessDefinitions):
    '''
    Methods for an individulal measurement within a graph
    
    Parameters
    ----------
    awrde: object variable
         The AWRDE object returned from awrde_utils.establish_link() 
    graph_name: string
         Name of the graph that the measurement belongs to
    measurement_name: string
         Full name of the measurement
    
    '''    
    def __init__(self, awrde, graph_name: str, measurement_name: str):#---------------------------------------
        self.awrde = awrde
        _ProcessDefinitions.__init__(self, awrde)
        self._initialize_measurement(graph_name, measurement_name)
        #
    def _initialize_measurement(self, graph_name, measurement_name):#---------------------------------
        try:
            self._gr = self.awrde.Project.Graphs(graph_name)
        except:
            raise RuntimeError('Graph does not exist: ' + graph_name)
        #end try
        #
        try:
            self._meas = self._gr.Measurements(measurement_name)
        except:
            raise RuntimeError(measurement_name + ' does not exist')
        #end try
        #
    @property
    def measurement_name(self) -> str:#-----------------------------------------------------
        '''Returns name of measurement'''
        return self._meas.Name
        #
    @property
    def measurement_enabled(self) -> bool:#---------------------------------------
        return self._meas.Enabled
        #
    @measurement_enabled.setter
    def measurement_enabled(self, enabled_state: bool):#----------------------------
        self._meas.Enabled = enabled_state
        #
    @property
    def measurement_axis(self):#----------------------------------------------------------
        return self._meas.AxisIndex, self._meas.OnLeftAxis
        #
    def set_measurement_axis(self, axis_index: int, on_left_axis: bool):#-------------------
        if not isinstance(axis_index, int):
            raise TypeError('axis_index must by type int')
        #end if
        if not isinstance(on_left_axis, bool):
            raise TypeError('on_left_axis must be type boolean')
        #end if
        try:
            self._meas.AxisIndex = axis_index
        except:
            raise RuntimeError('problem with setting axis index')
        #end try
        self._meas.OnLeftAxis = on_left_axis
        #
    @property
    def num_traces(self) -> int:#--------------------------------------------------
        ''' Returns number of traces in measurement'''
        return self._meas.TraceCount
        #
    @property
    def trace_data(self) -> list[np.ndarray]:#------------------------------------
        '''
        Returns a list of arrays. Each array corresponds to a measurement trace
        
        '''
        trace_data_list = list()
        if self.measurement_enabled:
            NumTraces = self.num_traces
            for trace_idx in range(1, NumTraces+1):
                trace_data_tuple = self._meas.TraceValues(trace_idx) #returns tuple
                trace_data_ay = np.asarray(trace_data_tuple) #convert tuple into array
                trace_data_list.append(trace_data_ay)
            #end for
            #
            trace_data_list = self._cleanup_vss_spectrum(trace_data_list)
            trace_data_list = self._scale_trace_data_to_project_units(trace_data_list)
        #end if
        #
        return trace_data_list
        #
    @property
    def sweep_var_labels(self) -> list:#-----------------------------------------------
        '''returns list of sweep variables'''
        sweep_param_dict = self._sweep_parameters
        sweep_var_list = list()
        for sweep_var_name in sweep_param_dict['SWPVAR Labels']:
            sweep_var_list.append(sweep_var_name)
        #end for
        return sweep_var_list
        #
    @property
    def sweep_var_trace_info(self) -> list:#--------------------------------------------
        '''returns list. Each list is a dictionary of sweep var parameters for each trace'''
        sweep_param_dict = self._sweep_parameters
        return sweep_param_dict['Trace Information']
        #
    @property
    def _sweep_parameters(self) -> dict:#-------------------------------------------------
        '''
        returns dictionary of SWPVAR items
        
        Returns
        -------
        sweep_param_dict: dictionary
               "Num SWPVARs" : Number of SWPVAR blocks detected
               "SWPVAR Labels" : list of strings. Each element is the sweep variable name
               "Trace Information" : list of dictionary of SWPVAR label:value for each trace.
        '''
        sweep_param_dict = {}
        num_traces = self.num_traces
        num_swpvars = self._meas.SweepLabels(1).Count
        sweep_param_dict['Num SWPVARs'] = num_swpvars
        #
        if num_swpvars > 0:
            swpvar_lbls_list = list()
            for lbl_idx in range(1, num_swpvars+1):
                swpvar_lbl = self._meas.SweepLabels(1).Item(lbl_idx).Name
                swpvar_lbls_list.append(swpvar_lbl)
            #end for
            sweep_param_dict['SWPVAR Labels'] = swpvar_lbls_list
            #
            trace_info_list = list()
            for trace_idx in range(1, num_traces+1):
                Trace_info_dict = {}
                for swpvar_idx in range(1, num_swpvars+1):
                    swpvar_name = self._meas.SweepLabels(trace_idx).Item(swpvar_idx).Name
                    swpvar_val = self._meas.SweepLabels(trace_idx).Item(swpvar_idx).Value
                    Trace_info_dict[swpvar_name] = swpvar_val
                #end for
                trace_info_list.append(Trace_info_dict)
            #end for
            sweep_param_dict['Trace Information'] = trace_info_list
            #
        else:
            sweep_param_dict['SWPVAR Labels'] = []
            sweep_param_dict['Trace Information'] = []
        #end if
        #
        return sweep_param_dict
        #
    def _scale_trace_data_to_project_units(self, trace_data_list: list) -> list:#-------------------------------
        '''
        measurement trace data read back in base units.
        This routine scales the data to project units.
        '''
        source_doc = self._meas.Source
        source_doc_type = self._source_doc_type(source_doc)
        if source_doc_type == 'Circuit Schematic':
            schem = self.awrde.Project.Schematics(source_doc)
            lpf_name = schem.ProcessDefinition
        elif source_doc_type == 'System Diagram':
            sys_diags = SystemDiagrams(self.awrde)
            lpf_name = sys_diags.Which_LPF
        #end if
        lpf_units_dict = self._project_units(lpf_name)
        #
        for axis_idx in range(2): #0=Xaxis, 1=Yaxis
            unit_type_enum = self._meas.UnitType(axis_idx+1)
            unit_type_str = self._unit_type_str(unit_type_enum)
            unit_string = lpf_units_dict[unit_type_str]
            #
            unit_scale_factor_dict = self._unit_scale_factor(unit_type_str, unit_string)
            offset1 = unit_scale_factor_dict['Offset1']
            offset2 = unit_scale_factor_dict['Offset2']
            multiplier = unit_scale_factor_dict['Multiplier']
            for trace_idx in range(len(trace_data_list)):
                trace_data_ay = trace_data_list[trace_idx]
                trace_data_ay[:,axis_idx] = (trace_data_ay[:,axis_idx] + offset2)*multiplier + offset1
                trace_data_list[trace_idx] = trace_data_ay
            #end for
        #end for
        return trace_data_list
        #
    def _cleanup_vss_spectrum(self, trace_data_list: list) -> list[np.ndarray]:#---------------------
        '''
        VSS time domain spectrum's last data point is some unusually large value.
        This routine strips off the last data point
        '''
        valid_cleanup_measurement = False
        measurement_type = self._meas.Type
        if 'V_SPEC' in measurement_type and (not 'RFI' in measurement_type):
            valid_cleanup_measurement = True
        elif ('PWR_SPEC' in measurement_type) and (not 'RFI' in measurement_type):
            valid_cleanup_measurement = True
        #end if
        #
        if valid_cleanup_measurement:
            temp_trace_data_list = copy.deepcopy(trace_data_list)
            trace_data_list = list()
            NumTraces = len(temp_trace_data_list)
            for trace_idx in range(NumTraces):
                trace_data_ay = temp_trace_data_list[trace_idx]
                trace_data_ay = trace_data_ay[:-1] #Remove last data point
                trace_data_list.append(trace_data_ay)
            #end for
        #end if
        return trace_data_list
        #
    def _source_doc_type(self, source_doc: str) -> str:#------------------------------------------------------
        ''' returns Source document type: System Diagram, Circuit Schematic'''
        source_doc_type = ''
        
        system_diags = _SystemDiagrams(self.awrde)
        system_diagram_name_list = system_diags.system_diagram_name_list
        #
        schematic_diagram_list = list()
        for schem in self.awrde.Project.Schematics:
            schematic_diagram_list.append(schem.Name)
        #end 
        #
        if source_doc in system_diagram_name_list:
            source_doc_type = 'System Diagram'
        elif source_doc in schematic_diagram_list:
            source_doc_type = 'Circuit Schematic'
        #end if
        return source_doc_type
        #
        
#
#*****************************************************************************
#
class _Axes():
    ''' 
    methods for all axes in the graph
    
    Parameters
    ----------
    awrde: object variable
         The AWRDE object returned from awrde_utils.establish_link() 
    
    graph_name: string
    
    '''
    def __init__(self, awrde, graph_name: str):#---------------------
        self.awrde = awrde
        self._initialize_axes(graph_name)
        self._build_axes_dict()
        #
    def _initialize_axes(self, graph_name):#------------------------------
        try:
            self._gr = self.awrde.Project.Graphs(graph_name)
        except:
            raise RuntimeError('Graph does not exist: ' + graph_name)
        #end try
        self._graph_name = graph_name
        #
    def _build_axes_dict(self):#-----------------------------------------
        ''' creates a dictionary of all the axes in the graph'''
        num_axes = self._gr.Axes.Count
        self._axes_dict = {}
        for ax_idx in range(1, num_axes+1):
            axis = _Axis(self.awrde, self._graph_name, ax_idx)
            self._axes_dict[axis.axis_name] = axis
        #end for
        #
    @property
    def axes_dict(self) -> dict:#-----------------------------------------
        '''returns axes dictionary'''
        return self._axes_dict
        #
    @property
    def axes_name_list(self) -> list:
        '''returns list of the axes names'''
        axes_name_list = list()
        for ax in self._gr.Axes:
            axes_name_list.append(ax.Name)
        #end for
        return axes_name_list
        #
#
#*****************************************************************************
#
class _Axis():
    '''
    Methods for an individulal measurement within a graph
    
    Parameters
    ----------
    awrde: object variable
         The AWRDE object returned from awrde_utils.establish_link()
    graph_name: string
         Name of the graph that the measurement belongs to
    axis_index: int
    '''
    def __init__(self, awrde, graph_name: str, axis_index: int):#-------------------
        self.awrde = awrde
        self._initialize_axis(graph_name, axis_index)
        #
    def _initialize_axis(self, graph_name, axis_index):#------------------------------
        if not isinstance(axis_index, int):
            raise TypeError('axis_index must be int type')
        #end if
        try:
            self._gr = self.awrde.Project.Graphs(graph_name)
        except:
            raise RuntimeError('Graph does not exist: ' + graph_name)
        #end try
        try:
            self._axis = self._gr.Axes(axis_index)
        except:
            raise RuntimeError('could not acces axis. Axis index='+str(axis_index))
        #end try
        #
    @property
    def axis_name(self) -> str: #---------------------------------------
        '''returns name of the axis'''
        return self._axis.Name
        #
    @property
    def axis_auto_scale(self) -> bool:#-------------------------------------------------------------
        return self._axis.AutoScale
        #
    @axis_auto_scale.setter
    def axis_auto_scale(self, auto_scale: bool):#------------------------------------------------
        if not isinstance(auto_scale, bool):
            raise TypeError('auto_scale must be boolean type')
        #end if
        self._axis.AutoScale = auto_scale
        #
    @property
    def axis_minimum_scale(self) -> float:#--------------------------------------------------
        return self._axis.MinimumScale
        #
    @axis_minimum_scale.setter
    def axis_minimum_scale(self, min_scale_val: float):#------------------------------------------
        self._axis.MinimumScale = min_scale_val
        #
    @property
    def axis_maximum_scale(self) -> float:#--------------------------------------------------
        return self._axis.MaximumScale
        #
    @axis_maximum_scale.setter
    def axis_maximum_scale(self, max_scale_val: float):#------------------------------------------
        self._axis.MaximumScale = max_scale_val
        #
    @property
    def axis_grid_step(self) -> float:#--------------------------------------------------------
        return self._axis.MajorGridlinesStep
        #
    @axis_grid_step.setter
    def axis_grid_step(self, step_size: float):#-------------------------------------------------
        self._axis.MajorGridlinesStep = step_size    
#
#*****************************************************************************
#
class _Markers():
    '''
    Methods for markers within a graph
    
    Parameters
    ----------
    awrde: object variable
         The AWRDE object returned from awrde_utils.establish_link() 
    graph_name: string
         Name of the graph that the measurement belongs to
    '''    
    def __init__(self, awrde, graph_name: str):#--------------------------------------------
        self.awrde = awrde
        self._initialize_markers(graph_name)
        self._build_markers_dict()
        #
    def _initialize_markers(self, graph_name):#---------------------------------------------
        try:
            self._gr = self.awrde.Project.Graphs(graph_name)
        except:
            raise RuntimeError('Graph does not exist: ' + graph_name)
        #end try
        self._graph_name = graph_name
        #
    def _build_markers_dict(self):#-------------------------------------------------
        '''creates a dictionary of all the markers in the graph'''
        self._markers_dict = {}
        marker_idx = 0
        for marker in self._gr.Markers:
            marker = _Marker(self.awrde, self._graph_name, marker.Name)
            self._markers_dict[marker_idx] = marker
            marker_idx += 1
        #end for
        #
    @property
    def markers_dict(self) -> dict:#-------------------------------------------
        '''returns markers dictionary'''
        self._build_markers_dict()
        return self._markers_dict
        #
    @property
    def marker_name_list(self) -> list[str]:#--------------------------------------------------
        '''
        Creates list of all the marker names in the graph
        
        Returns
        -------
        marker_name_list: list[string]
                  Each item in the list is a marker name
        '''
        marker_name_list = list()
        for mkr in self._gr.Markers:
            marker_name_list.append(mkr.Name)
        #end for
        return marker_name_list
        #
    @property
    def remove_all_markers(self):#-----------------------------------------------------------
        self._gr.Markers.RemoveAll()
        #
    def remove_marker(self, marker_name: str) -> bool:#--------------------------------------
        marker_removed = False
        try:
            for mkr_idx in range(1, self._gr.Markers.Count+1):
                if self._gr.Markers(mkr_idx).Name == marker_name:
                    self._gr.Markers.Remove(mkr_idx)
                    marker_removed = True
                    break
                #end if
            #end if
        except:
            warnings.warn('Could not remove marker')
        #end try
        return marker_removed
        #
    def add_marker(self, measurement: str, sweep_value, trace_index: int=0):#---------
        if not isinstance(measurement, str):
            raise TypeError('measurement must be string type')
        #end if
        if not isinstance(sweep_value, float) and not isinstance(sweep_value,int):
            raise TypeError('sweep_value must be int or float type')
        #
        if not isinstance(trace_index, int):
            raise TypeError('trace_index must be int type')
        #end
        try:
            self._gr.Markers.AddEx(measurement, 1, sweep_value, trace_index + 1)
        except:
            raise RuntimeError('Could not add marker to graph')
        #end if
        #      
#
#*****************************************************************************
#
class _Marker():
    '''
    Methods for an dividual marker
    
    Parameters
    ----------
    awrde: object variable
         The AWRDE object returned from awrde_utils.establish_link() 
    graph_name: string
         Name of the graph that the measurement belongs to
    marker_name: string
    '''        
    def __init__(self, awrde, graph_name, marker_name):#----------------------------------------
        self.awrde = awrde
        self._initialize_marker(graph_name, marker_name)
        #
    def _initialize_marker(self, graph_name, marker_name):#---------------------------------
        try:
            self._gr = self.awrde.Project.Graphs(graph_name)
        except:
            raise RuntimeError('Graph does not exist: ' + graph_name)
        #end try
        #
        try:
            self._mkr = self._gr.Markers(marker_name)
        except:
            raise RuntimeError('marker not recognized: ', marker_name)
        #end try
        #       
    @property
    def marker_name(self) -> str:#----------------------------------------------------------------
        return self._mkr.Name     
        #
    @property
    def marker_measurement(self) -> str:#----------------------------------------------------------
        return self._mkr.Measurement
        #
    @property
    def marker_data_text(self) -> str:#--------------------------------------------------------
        return self._mkr.DataValueText
        #
    @property
    def marker_data_value_x(self) -> float:#----------------------------------------------------
        return self._mkr.DataValue(1)
        #
    @property
    def marker_data_value_y1(self) -> float:#----------------------------------------------------
        return self._mkr.DataValue(2)
        #    
    @property
    def marker_data_value_y2(self) -> float:#----------------------------------------------------
        try:
            mkr_data_value = self._mkr.DataValue(3)
        except:
            raise RuntimeError('marker Y2 invalid')
        #end try
        return mkr_data_value
        #
    @property
    def marker_sweep_value(self) -> float:#-------------------------------------------------------
        return self._mkr.SweepValue
        #
    @marker_sweep_value.setter
    def marker_sweep_value(self, x_axis_value: float):#----------------------------------------------
        self._mkr.SweepValue = x_axis_value
        #
    @property
    def trace_index(self) -> int:#-----------------------------------------------------------
        #0 base indexing
        return int(self._mkr.TraceIndex - 1)
        #
    @trace_index.setter
    def trace_index(self, trace_index: int):#-----------------------------------------------
        try:
            self._mkr.TraceIndex =  trace_index + 1
        except:
            pass #let it lie
        #end try
        #
    @property
    def marker_to_max(self):#-----------------------------------------------------------------
        self._mkr.MoveToMaximum()
        #
    @property
    def marker_to_min(self):#-----------------------------------------------------------------
        self._mkr.MoveToMinimum()
        #
    
#
#*****************************************************************************
#
class _Graph(_Measurements, _Axes, _Markers):
    '''
    Methods for an individual graph in the AWRDE project
    
    Parameters
    ----------
    awrde: object variable
         The AWRDE object returned from awrde_utils.establish_link()
    
    Graph name: string
          name of an existing graph in the AWRDE project
    
    '''
    def __init__(self, awrde, graph_name: str):#---------------------------------------------
        self.awrde = awrde
        _Axes.__init__(self, awrde, graph_name)
        _Measurements.__init__(self, awrde, graph_name)
        _Markers.__init__(self, awrde, graph_name)
        self._initialize_graph(graph_name)
        #
    def _initialize_graph(self, graph_name):#-------------------------------------------------------
        try:
            self._gr = self.awrde.Project.Graphs(graph_name)
        except:
            raise RuntimeError('Graph does not exist: ' + graph_name)
        #end try
        #
    @property
    def graph_name(self) -> str:#----------------------------------------------------------------
        '''Returns name of Graph'''
        return self._gr.Name
        #
    @property
    def graph_type(self) -> str: #----------------------------------------------------------
        '''
        Reports graph type 
        
        Returns
        -------
        graph_type_str: string
              Rectangular, Rectangular-Real/Imag, Smith Chart, Polar, Histogram, Antenna Plot,
              Tabular, Constellation, 3D Plot
        
        '''
        graph_type_enum = self._gr.Type
        if mwo.mwGraphType(graph_type_enum) == mwo.mwGraphType.mwGT_Rectangular:
            graph_type_str = 'Rectangular'
        elif graph_type_enum == mwo.mwGraphType.mwGT_RectangularComplex:
            graph_type_str = 'Rectangular-Real/Imag'
        elif mwo.mwGraphType(graph_type_enum) == mwo.mwGraphType.mwGT_SmithChart:
            graph_type_str = 'Smith Chart'
        elif graph_type_enum == mwo.mwGraphType.mwGT_Polar:
            graph_type_str = 'Polar'
        elif mwo.mwGraphType(graph_type_enum) == mwo.mwGraphType.mwGT_Histogram:
            graph_type_str = 'Histogram'
        elif mwo.mwGraphType(graph_type_enum) == mwo.mwGraphType.mwGT_Antenna:
            graph_type_str = 'Antenna Plot'
        elif mwo.mwGraphType(graph_type_enum) == mwo.mwGraphType.mwGT_Tabular:
            graph_type_str = 'Tabular'
        elif mwo.mwGraphType(graph_type_enum) == mwo.mwGraphType.mwGT_Constellation:
            graph_type_str = 'Constellation'
        elif mwo.mwGraphType(graph_type_enum) == mwo.mwGraphType.mwGT_ThreeDim:
            graph_type_str = '3D Plot'
        else:
            warnings.warn('Graph Type cannot be determined')
            graph_type_str = '???'
        #end if
        return graph_type_str
        #
    def freeze_traces(self, freeze_traces: bool):#-----------------------------------------------------------------
        '''Sets Freeze Traces or Clear Frozen'''
        if not isinstance(freeze_traces,bool):
            raise TypeError('freeze_trace_cmd must be boolean')
        #end if
        if freeze_traces == True:
            self._gr.FreezeTraces()
        else:
            self._gr.ClearFrozenTraces()
        #end if
        #       